//
// Generator of src/frontend/cxx-gccbuiltin-sse.h for gcc
//
// Use make generate_builtins_ia32 (at the tp level) to compile this file
//

#include <iostream>
#include <sstream>
#include "builtins-common.hpp"

// --------------------------------------------
// Specific generators for NEON
// --------------------------------------------

#include <stdint.h>

#define GENERATE_NEON_VECTOR(BITS, T) \
template <> \
struct generate_type<__simd##BITS##_##T> \
{ \
    static const int size = BITS / 8;\
    typedef T element_type;\
    static std::string g() \
    {\
        std::stringstream ss;\
\
        ss << "get_vector_type_by_bytes(" << generate_type<element_type>::g() << ", " << (size) << ")";\
\
        return ss.str();\
    }\
};

template <>
struct generate_type<__builtin_neon_ti> // 2 double registers
{
    static const int size = 2;
    typedef long long int element_type;

    static std::string g()
    {
        std::stringstream ss;
        ss << "get_vector_type_by_elements(" << generate_type<element_type>::g() << ", " << (size) << ")";\
        return ss.str();
    }
};

template <>
struct generate_type<__builtin_neon_ei> // 3 double registers
{
    static const int size = 3;
    typedef long long int element_type;

    static std::string g()
    {
        std::stringstream ss;
        ss << "get_vector_type_by_elements(" << generate_type<element_type>::g() << ", " << (size) << ")";\
        return ss.str();
    }
};

template <>
struct generate_type<__builtin_neon_oi> // 4 double registers
{
    static const int size = 4;
    typedef long long int element_type;

    static std::string g()
    {
        std::stringstream ss;
        ss << "get_vector_type_by_elements(" << generate_type<element_type>::g() << ", " << (size) << ")";\
        return ss.str();
    }
};

template <>
struct generate_type<__builtin_neon_ci> // 6 double registers
{
    static const int size = 6;
    typedef long long int element_type;

    static std::string g()
    {
        std::stringstream ss;
        ss << "get_vector_type_by_elements(" << generate_type<element_type>::g() << ", " << (size) << ")";\
        return ss.str();
    }
};

template <>
struct generate_type<__builtin_neon_xi> // 8 double registers
{
    static const int size = 8;
    typedef long long int element_type;

    static std::string g()
    {
        std::stringstream ss;
        ss << "get_vector_type_by_elements(" << generate_type<element_type>::g() << ", " << (size) << ")";\
        return ss.str();
    }
};

typedef float float32_t;

GENERATE_NEON_VECTOR(128, float32_t)
GENERATE_NEON_VECTOR(128, int16_t)
GENERATE_NEON_VECTOR(128, int32_t)
GENERATE_NEON_VECTOR(128, int64_t)
GENERATE_NEON_VECTOR(128, int8_t)
GENERATE_NEON_VECTOR(64, float32_t)
GENERATE_NEON_VECTOR(64, int16_t)
GENERATE_NEON_VECTOR(64, int32_t)
GENERATE_NEON_VECTOR(64, int8_t)


// --------------------------------------------
// End of specific generators for NEON
// --------------------------------------------


#define END

#define VECTOR_INTRINSICS_LIST \
VECTOR_INTRIN(__builtin_neon_vabalsv2si) \
VECTOR_INTRIN(__builtin_neon_vabalsv4hi) \
VECTOR_INTRIN(__builtin_neon_vabalsv8qi) \
VECTOR_INTRIN(__builtin_neon_vabaluv2si) \
VECTOR_INTRIN(__builtin_neon_vabaluv4hi) \
VECTOR_INTRIN(__builtin_neon_vabaluv8qi) \
VECTOR_INTRIN(__builtin_neon_vabasv16qi) \
VECTOR_INTRIN(__builtin_neon_vabasv2si) \
VECTOR_INTRIN(__builtin_neon_vabasv4hi) \
VECTOR_INTRIN(__builtin_neon_vabasv4si) \
VECTOR_INTRIN(__builtin_neon_vabasv8hi) \
VECTOR_INTRIN(__builtin_neon_vabasv8qi) \
VECTOR_INTRIN(__builtin_neon_vabauv16qi) \
VECTOR_INTRIN(__builtin_neon_vabauv2si) \
VECTOR_INTRIN(__builtin_neon_vabauv4hi) \
VECTOR_INTRIN(__builtin_neon_vabauv4si) \
VECTOR_INTRIN(__builtin_neon_vabauv8hi) \
VECTOR_INTRIN(__builtin_neon_vabauv8qi) \
VECTOR_INTRIN(__builtin_neon_vabdfv2sf) \
VECTOR_INTRIN(__builtin_neon_vabdfv4sf) \
VECTOR_INTRIN(__builtin_neon_vabdlsv2si) \
VECTOR_INTRIN(__builtin_neon_vabdlsv4hi) \
VECTOR_INTRIN(__builtin_neon_vabdlsv8qi) \
VECTOR_INTRIN(__builtin_neon_vabdluv2si) \
VECTOR_INTRIN(__builtin_neon_vabdluv4hi) \
VECTOR_INTRIN(__builtin_neon_vabdluv8qi) \
VECTOR_INTRIN(__builtin_neon_vabdsv16qi) \
VECTOR_INTRIN(__builtin_neon_vabdsv2si) \
VECTOR_INTRIN(__builtin_neon_vabdsv4hi) \
VECTOR_INTRIN(__builtin_neon_vabdsv4si) \
VECTOR_INTRIN(__builtin_neon_vabdsv8hi) \
VECTOR_INTRIN(__builtin_neon_vabdsv8qi) \
VECTOR_INTRIN(__builtin_neon_vabduv16qi) \
VECTOR_INTRIN(__builtin_neon_vabduv2si) \
VECTOR_INTRIN(__builtin_neon_vabduv4hi) \
VECTOR_INTRIN(__builtin_neon_vabduv4si) \
VECTOR_INTRIN(__builtin_neon_vabduv8hi) \
VECTOR_INTRIN(__builtin_neon_vabduv8qi) \
VECTOR_INTRIN(__builtin_neon_vabsv16qi) \
VECTOR_INTRIN(__builtin_neon_vabsv2sf) \
VECTOR_INTRIN(__builtin_neon_vabsv2si) \
VECTOR_INTRIN(__builtin_neon_vabsv4hi) \
VECTOR_INTRIN(__builtin_neon_vabsv4sf) \
VECTOR_INTRIN(__builtin_neon_vabsv4si) \
VECTOR_INTRIN(__builtin_neon_vabsv8hi) \
VECTOR_INTRIN(__builtin_neon_vabsv8qi) \
VECTOR_INTRIN(__builtin_neon_vaddhnv2di) \
VECTOR_INTRIN(__builtin_neon_vaddhnv4si) \
VECTOR_INTRIN(__builtin_neon_vaddhnv8hi) \
VECTOR_INTRIN(__builtin_neon_vaddlsv2si) \
VECTOR_INTRIN(__builtin_neon_vaddlsv4hi) \
VECTOR_INTRIN(__builtin_neon_vaddlsv8qi) \
VECTOR_INTRIN(__builtin_neon_vaddluv2si) \
VECTOR_INTRIN(__builtin_neon_vaddluv4hi) \
VECTOR_INTRIN(__builtin_neon_vaddluv8qi) \
VECTOR_INTRIN(__builtin_neon_vaddv2sf) \
VECTOR_INTRIN(__builtin_neon_vaddv4sf) \
VECTOR_INTRIN(__builtin_neon_vaddwsv2si) \
VECTOR_INTRIN(__builtin_neon_vaddwsv4hi) \
VECTOR_INTRIN(__builtin_neon_vaddwsv8qi) \
VECTOR_INTRIN(__builtin_neon_vaddwuv2si) \
VECTOR_INTRIN(__builtin_neon_vaddwuv4hi) \
VECTOR_INTRIN(__builtin_neon_vaddwuv8qi) \
VECTOR_INTRIN(__builtin_neon_vbsldi) \
VECTOR_INTRIN(__builtin_neon_vbslv16qi) \
VECTOR_INTRIN(__builtin_neon_vbslv2di) \
VECTOR_INTRIN(__builtin_neon_vbslv2sf) \
VECTOR_INTRIN(__builtin_neon_vbslv2si) \
VECTOR_INTRIN(__builtin_neon_vbslv4hi) \
VECTOR_INTRIN(__builtin_neon_vbslv4sf) \
VECTOR_INTRIN(__builtin_neon_vbslv4si) \
VECTOR_INTRIN(__builtin_neon_vbslv8hi) \
VECTOR_INTRIN(__builtin_neon_vbslv8qi) \
VECTOR_INTRIN(__builtin_neon_vcagev2sf) \
VECTOR_INTRIN(__builtin_neon_vcagev4sf) \
VECTOR_INTRIN(__builtin_neon_vcagtv2sf) \
VECTOR_INTRIN(__builtin_neon_vcagtv4sf) \
VECTOR_INTRIN(__builtin_neon_vceqv16qi) \
VECTOR_INTRIN(__builtin_neon_vceqv2sf) \
VECTOR_INTRIN(__builtin_neon_vceqv2si) \
VECTOR_INTRIN(__builtin_neon_vceqv4hi) \
VECTOR_INTRIN(__builtin_neon_vceqv4sf) \
VECTOR_INTRIN(__builtin_neon_vceqv4si) \
VECTOR_INTRIN(__builtin_neon_vceqv8hi) \
VECTOR_INTRIN(__builtin_neon_vceqv8qi) \
VECTOR_INTRIN(__builtin_neon_vcgeuv16qi) \
VECTOR_INTRIN(__builtin_neon_vcgeuv2si) \
VECTOR_INTRIN(__builtin_neon_vcgeuv4hi) \
VECTOR_INTRIN(__builtin_neon_vcgeuv4si) \
VECTOR_INTRIN(__builtin_neon_vcgeuv8hi) \
VECTOR_INTRIN(__builtin_neon_vcgeuv8qi) \
VECTOR_INTRIN(__builtin_neon_vcgev16qi) \
VECTOR_INTRIN(__builtin_neon_vcgev2sf) \
VECTOR_INTRIN(__builtin_neon_vcgev2si) \
VECTOR_INTRIN(__builtin_neon_vcgev4hi) \
VECTOR_INTRIN(__builtin_neon_vcgev4sf) \
VECTOR_INTRIN(__builtin_neon_vcgev4si) \
VECTOR_INTRIN(__builtin_neon_vcgev8hi) \
VECTOR_INTRIN(__builtin_neon_vcgev8qi) \
VECTOR_INTRIN(__builtin_neon_vcgtuv16qi) \
VECTOR_INTRIN(__builtin_neon_vcgtuv2si) \
VECTOR_INTRIN(__builtin_neon_vcgtuv4hi) \
VECTOR_INTRIN(__builtin_neon_vcgtuv4si) \
VECTOR_INTRIN(__builtin_neon_vcgtuv8hi) \
VECTOR_INTRIN(__builtin_neon_vcgtuv8qi) \
VECTOR_INTRIN(__builtin_neon_vcgtv16qi) \
VECTOR_INTRIN(__builtin_neon_vcgtv2sf) \
VECTOR_INTRIN(__builtin_neon_vcgtv2si) \
VECTOR_INTRIN(__builtin_neon_vcgtv4hi) \
VECTOR_INTRIN(__builtin_neon_vcgtv4sf) \
VECTOR_INTRIN(__builtin_neon_vcgtv4si) \
VECTOR_INTRIN(__builtin_neon_vcgtv8hi) \
VECTOR_INTRIN(__builtin_neon_vcgtv8qi) \
VECTOR_INTRIN(__builtin_neon_vclsv16qi) \
VECTOR_INTRIN(__builtin_neon_vclsv2si) \
VECTOR_INTRIN(__builtin_neon_vclsv4hi) \
VECTOR_INTRIN(__builtin_neon_vclsv4si) \
VECTOR_INTRIN(__builtin_neon_vclsv8hi) \
VECTOR_INTRIN(__builtin_neon_vclsv8qi) \
VECTOR_INTRIN(__builtin_neon_vclzv16qi) \
VECTOR_INTRIN(__builtin_neon_vclzv2si) \
VECTOR_INTRIN(__builtin_neon_vclzv4hi) \
VECTOR_INTRIN(__builtin_neon_vclzv4si) \
VECTOR_INTRIN(__builtin_neon_vclzv8hi) \
VECTOR_INTRIN(__builtin_neon_vclzv8qi) \
VECTOR_INTRIN(__builtin_neon_vcntv16qi) \
VECTOR_INTRIN(__builtin_neon_vcntv8qi) \
VECTOR_INTRIN(__builtin_neon_vcombinedi) \
VECTOR_INTRIN(__builtin_neon_vcombinev2sf) \
VECTOR_INTRIN(__builtin_neon_vcombinev2si) \
VECTOR_INTRIN(__builtin_neon_vcombinev4hi) \
VECTOR_INTRIN(__builtin_neon_vcombinev8qi) \
VECTOR_INTRIN(__builtin_neon_vcreatedi) \
VECTOR_INTRIN(__builtin_neon_vcreatev2sf) \
VECTOR_INTRIN(__builtin_neon_vcreatev2si) \
VECTOR_INTRIN(__builtin_neon_vcreatev4hi) \
VECTOR_INTRIN(__builtin_neon_vcreatev8qi) \
VECTOR_INTRIN(__builtin_neon_vcvts_nv2sf) \
VECTOR_INTRIN(__builtin_neon_vcvts_nv2si) \
VECTOR_INTRIN(__builtin_neon_vcvts_nv4sf) \
VECTOR_INTRIN(__builtin_neon_vcvts_nv4si) \
VECTOR_INTRIN(__builtin_neon_vcvtsv2sf) \
VECTOR_INTRIN(__builtin_neon_vcvtsv2si) \
VECTOR_INTRIN(__builtin_neon_vcvtsv4sf) \
VECTOR_INTRIN(__builtin_neon_vcvtsv4si) \
VECTOR_INTRIN(__builtin_neon_vcvtu_nv2sf) \
VECTOR_INTRIN(__builtin_neon_vcvtu_nv2si) \
VECTOR_INTRIN(__builtin_neon_vcvtu_nv4sf) \
VECTOR_INTRIN(__builtin_neon_vcvtu_nv4si) \
VECTOR_INTRIN(__builtin_neon_vcvtuv2sf) \
VECTOR_INTRIN(__builtin_neon_vcvtuv2si) \
VECTOR_INTRIN(__builtin_neon_vcvtuv4sf) \
VECTOR_INTRIN(__builtin_neon_vcvtuv4si) \
VECTOR_INTRIN(__builtin_neon_vdup_lanedi) \
VECTOR_INTRIN(__builtin_neon_vdup_lanev16qi) \
VECTOR_INTRIN(__builtin_neon_vdup_lanev2di) \
VECTOR_INTRIN(__builtin_neon_vdup_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vdup_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vdup_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vdup_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vdup_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vdup_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vdup_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vdup_ndi) \
VECTOR_INTRIN(__builtin_neon_vdup_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vdup_nv2di) \
VECTOR_INTRIN(__builtin_neon_vdup_nv2sf) \
VECTOR_INTRIN(__builtin_neon_vdup_nv2si) \
VECTOR_INTRIN(__builtin_neon_vdup_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vdup_nv4sf) \
VECTOR_INTRIN(__builtin_neon_vdup_nv4si) \
VECTOR_INTRIN(__builtin_neon_vdup_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vdup_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vextdi) \
VECTOR_INTRIN(__builtin_neon_vextv16qi) \
VECTOR_INTRIN(__builtin_neon_vextv2di) \
VECTOR_INTRIN(__builtin_neon_vextv2sf) \
VECTOR_INTRIN(__builtin_neon_vextv2si) \
VECTOR_INTRIN(__builtin_neon_vextv4hi) \
VECTOR_INTRIN(__builtin_neon_vextv4sf) \
VECTOR_INTRIN(__builtin_neon_vextv4si) \
VECTOR_INTRIN(__builtin_neon_vextv8hi) \
VECTOR_INTRIN(__builtin_neon_vextv8qi) \
VECTOR_INTRIN(__builtin_neon_vget_highv16qi) \
VECTOR_INTRIN(__builtin_neon_vget_highv2di) \
VECTOR_INTRIN(__builtin_neon_vget_highv4sf) \
VECTOR_INTRIN(__builtin_neon_vget_highv4si) \
VECTOR_INTRIN(__builtin_neon_vget_highv8hi) \
VECTOR_INTRIN(__builtin_neon_vget_lanedi) \
VECTOR_INTRIN(__builtin_neon_vget_laneuv16qi) \
VECTOR_INTRIN(__builtin_neon_vget_laneuv2si) \
VECTOR_INTRIN(__builtin_neon_vget_laneuv4hi) \
VECTOR_INTRIN(__builtin_neon_vget_laneuv4si) \
VECTOR_INTRIN(__builtin_neon_vget_laneuv8hi) \
VECTOR_INTRIN(__builtin_neon_vget_laneuv8qi) \
VECTOR_INTRIN(__builtin_neon_vget_lanev16qi) \
VECTOR_INTRIN(__builtin_neon_vget_lanev2di) \
VECTOR_INTRIN(__builtin_neon_vget_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vget_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vget_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vget_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vget_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vget_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vget_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vget_lowv16qi) \
VECTOR_INTRIN(__builtin_neon_vget_lowv2di) \
VECTOR_INTRIN(__builtin_neon_vget_lowv4sf) \
VECTOR_INTRIN(__builtin_neon_vget_lowv4si) \
VECTOR_INTRIN(__builtin_neon_vget_lowv8hi) \
VECTOR_INTRIN(__builtin_neon_vhaddsv16qi) \
VECTOR_INTRIN(__builtin_neon_vhaddsv2si) \
VECTOR_INTRIN(__builtin_neon_vhaddsv4hi) \
VECTOR_INTRIN(__builtin_neon_vhaddsv4si) \
VECTOR_INTRIN(__builtin_neon_vhaddsv8hi) \
VECTOR_INTRIN(__builtin_neon_vhaddsv8qi) \
VECTOR_INTRIN(__builtin_neon_vhadduv16qi) \
VECTOR_INTRIN(__builtin_neon_vhadduv2si) \
VECTOR_INTRIN(__builtin_neon_vhadduv4hi) \
VECTOR_INTRIN(__builtin_neon_vhadduv4si) \
VECTOR_INTRIN(__builtin_neon_vhadduv8hi) \
VECTOR_INTRIN(__builtin_neon_vhadduv8qi) \
VECTOR_INTRIN(__builtin_neon_vhsubsv16qi) \
VECTOR_INTRIN(__builtin_neon_vhsubsv2si) \
VECTOR_INTRIN(__builtin_neon_vhsubsv4hi) \
VECTOR_INTRIN(__builtin_neon_vhsubsv4si) \
VECTOR_INTRIN(__builtin_neon_vhsubsv8hi) \
VECTOR_INTRIN(__builtin_neon_vhsubsv8qi) \
VECTOR_INTRIN(__builtin_neon_vhsubuv16qi) \
VECTOR_INTRIN(__builtin_neon_vhsubuv2si) \
VECTOR_INTRIN(__builtin_neon_vhsubuv4hi) \
VECTOR_INTRIN(__builtin_neon_vhsubuv4si) \
VECTOR_INTRIN(__builtin_neon_vhsubuv8hi) \
VECTOR_INTRIN(__builtin_neon_vhsubuv8qi) \
VECTOR_INTRIN(__builtin_neon_vld1di) \
VECTOR_INTRIN(__builtin_neon_vld1_dupdi) \
VECTOR_INTRIN(__builtin_neon_vld1_dupv16qi) \
VECTOR_INTRIN(__builtin_neon_vld1_dupv2di) \
VECTOR_INTRIN(__builtin_neon_vld1_dupv2sf) \
VECTOR_INTRIN(__builtin_neon_vld1_dupv2si) \
VECTOR_INTRIN(__builtin_neon_vld1_dupv4hi) \
VECTOR_INTRIN(__builtin_neon_vld1_dupv4sf) \
VECTOR_INTRIN(__builtin_neon_vld1_dupv4si) \
VECTOR_INTRIN(__builtin_neon_vld1_dupv8hi) \
VECTOR_INTRIN(__builtin_neon_vld1_dupv8qi) \
VECTOR_INTRIN(__builtin_neon_vld1_lanedi) \
VECTOR_INTRIN(__builtin_neon_vld1_lanev16qi) \
VECTOR_INTRIN(__builtin_neon_vld1_lanev2di) \
VECTOR_INTRIN(__builtin_neon_vld1_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vld1_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vld1_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vld1_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vld1_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vld1_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vld1_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vld1v16qi) \
VECTOR_INTRIN(__builtin_neon_vld1v2di) \
VECTOR_INTRIN(__builtin_neon_vld1v2sf) \
VECTOR_INTRIN(__builtin_neon_vld1v2si) \
VECTOR_INTRIN(__builtin_neon_vld1v4hi) \
VECTOR_INTRIN(__builtin_neon_vld1v4sf) \
VECTOR_INTRIN(__builtin_neon_vld1v4si) \
VECTOR_INTRIN(__builtin_neon_vld1v8hi) \
VECTOR_INTRIN(__builtin_neon_vld1v8qi) \
VECTOR_INTRIN(__builtin_neon_vld2di) \
VECTOR_INTRIN(__builtin_neon_vld2_dupdi) \
VECTOR_INTRIN(__builtin_neon_vld2_dupv2sf) \
VECTOR_INTRIN(__builtin_neon_vld2_dupv2si) \
VECTOR_INTRIN(__builtin_neon_vld2_dupv4hi) \
VECTOR_INTRIN(__builtin_neon_vld2_dupv8qi) \
VECTOR_INTRIN(__builtin_neon_vld2_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vld2_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vld2_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vld2_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vld2_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vld2_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vld2_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vld2v16qi) \
VECTOR_INTRIN(__builtin_neon_vld2v2sf) \
VECTOR_INTRIN(__builtin_neon_vld2v2si) \
VECTOR_INTRIN(__builtin_neon_vld2v4hi) \
VECTOR_INTRIN(__builtin_neon_vld2v4sf) \
VECTOR_INTRIN(__builtin_neon_vld2v4si) \
VECTOR_INTRIN(__builtin_neon_vld2v8hi) \
VECTOR_INTRIN(__builtin_neon_vld2v8qi) \
VECTOR_INTRIN(__builtin_neon_vld3di) \
VECTOR_INTRIN(__builtin_neon_vld3_dupdi) \
VECTOR_INTRIN(__builtin_neon_vld3_dupv2sf) \
VECTOR_INTRIN(__builtin_neon_vld3_dupv2si) \
VECTOR_INTRIN(__builtin_neon_vld3_dupv4hi) \
VECTOR_INTRIN(__builtin_neon_vld3_dupv8qi) \
VECTOR_INTRIN(__builtin_neon_vld3_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vld3_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vld3_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vld3_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vld3_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vld3_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vld3_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vld3v16qi) \
VECTOR_INTRIN(__builtin_neon_vld3v2sf) \
VECTOR_INTRIN(__builtin_neon_vld3v2si) \
VECTOR_INTRIN(__builtin_neon_vld3v4hi) \
VECTOR_INTRIN(__builtin_neon_vld3v4sf) \
VECTOR_INTRIN(__builtin_neon_vld3v4si) \
VECTOR_INTRIN(__builtin_neon_vld3v8hi) \
VECTOR_INTRIN(__builtin_neon_vld3v8qi) \
VECTOR_INTRIN(__builtin_neon_vld4di) \
VECTOR_INTRIN(__builtin_neon_vld4_dupdi) \
VECTOR_INTRIN(__builtin_neon_vld4_dupv2sf) \
VECTOR_INTRIN(__builtin_neon_vld4_dupv2si) \
VECTOR_INTRIN(__builtin_neon_vld4_dupv4hi) \
VECTOR_INTRIN(__builtin_neon_vld4_dupv8qi) \
VECTOR_INTRIN(__builtin_neon_vld4_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vld4_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vld4_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vld4_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vld4_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vld4_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vld4_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vld4v16qi) \
VECTOR_INTRIN(__builtin_neon_vld4v2sf) \
VECTOR_INTRIN(__builtin_neon_vld4v2si) \
VECTOR_INTRIN(__builtin_neon_vld4v4hi) \
VECTOR_INTRIN(__builtin_neon_vld4v4sf) \
VECTOR_INTRIN(__builtin_neon_vld4v4si) \
VECTOR_INTRIN(__builtin_neon_vld4v8hi) \
VECTOR_INTRIN(__builtin_neon_vld4v8qi) \
VECTOR_INTRIN(__builtin_neon_vmaxfv2sf) \
VECTOR_INTRIN(__builtin_neon_vmaxfv4sf) \
VECTOR_INTRIN(__builtin_neon_vmaxsv16qi) \
VECTOR_INTRIN(__builtin_neon_vmaxsv2si) \
VECTOR_INTRIN(__builtin_neon_vmaxsv4hi) \
VECTOR_INTRIN(__builtin_neon_vmaxsv4si) \
VECTOR_INTRIN(__builtin_neon_vmaxsv8hi) \
VECTOR_INTRIN(__builtin_neon_vmaxsv8qi) \
VECTOR_INTRIN(__builtin_neon_vmaxuv16qi) \
VECTOR_INTRIN(__builtin_neon_vmaxuv2si) \
VECTOR_INTRIN(__builtin_neon_vmaxuv4hi) \
VECTOR_INTRIN(__builtin_neon_vmaxuv4si) \
VECTOR_INTRIN(__builtin_neon_vmaxuv8hi) \
VECTOR_INTRIN(__builtin_neon_vmaxuv8qi) \
VECTOR_INTRIN(__builtin_neon_vminfv2sf) \
VECTOR_INTRIN(__builtin_neon_vminfv4sf) \
VECTOR_INTRIN(__builtin_neon_vminsv16qi) \
VECTOR_INTRIN(__builtin_neon_vminsv2si) \
VECTOR_INTRIN(__builtin_neon_vminsv4hi) \
VECTOR_INTRIN(__builtin_neon_vminsv4si) \
VECTOR_INTRIN(__builtin_neon_vminsv8hi) \
VECTOR_INTRIN(__builtin_neon_vminsv8qi) \
VECTOR_INTRIN(__builtin_neon_vminuv16qi) \
VECTOR_INTRIN(__builtin_neon_vminuv2si) \
VECTOR_INTRIN(__builtin_neon_vminuv4hi) \
VECTOR_INTRIN(__builtin_neon_vminuv4si) \
VECTOR_INTRIN(__builtin_neon_vminuv8hi) \
VECTOR_INTRIN(__builtin_neon_vminuv8qi) \
VECTOR_INTRIN(__builtin_neon_vmla_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vmla_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vmla_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vmla_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vmla_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vmla_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vmlals_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vmlals_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vmlals_nv2si) \
VECTOR_INTRIN(__builtin_neon_vmlals_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vmlalsv2si) \
VECTOR_INTRIN(__builtin_neon_vmlalsv4hi) \
VECTOR_INTRIN(__builtin_neon_vmlalsv8qi) \
VECTOR_INTRIN(__builtin_neon_vmlalu_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vmlalu_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vmlalu_nv2si) \
VECTOR_INTRIN(__builtin_neon_vmlalu_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vmlaluv2si) \
VECTOR_INTRIN(__builtin_neon_vmlaluv4hi) \
VECTOR_INTRIN(__builtin_neon_vmlaluv8qi) \
VECTOR_INTRIN(__builtin_neon_vmla_nv2sf) \
VECTOR_INTRIN(__builtin_neon_vmla_nv2si) \
VECTOR_INTRIN(__builtin_neon_vmla_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vmla_nv4sf) \
VECTOR_INTRIN(__builtin_neon_vmla_nv4si) \
VECTOR_INTRIN(__builtin_neon_vmla_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vmlav16qi) \
VECTOR_INTRIN(__builtin_neon_vmlav2sf) \
VECTOR_INTRIN(__builtin_neon_vmlav2si) \
VECTOR_INTRIN(__builtin_neon_vmlav4hi) \
VECTOR_INTRIN(__builtin_neon_vmlav4sf) \
VECTOR_INTRIN(__builtin_neon_vmlav4si) \
VECTOR_INTRIN(__builtin_neon_vmlav8hi) \
VECTOR_INTRIN(__builtin_neon_vmlav8qi) \
VECTOR_INTRIN(__builtin_neon_vmls_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vmls_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vmls_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vmls_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vmls_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vmls_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vmlsls_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vmlsls_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vmlsls_nv2si) \
VECTOR_INTRIN(__builtin_neon_vmlsls_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vmlslsv2si) \
VECTOR_INTRIN(__builtin_neon_vmlslsv4hi) \
VECTOR_INTRIN(__builtin_neon_vmlslsv8qi) \
VECTOR_INTRIN(__builtin_neon_vmlslu_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vmlslu_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vmlslu_nv2si) \
VECTOR_INTRIN(__builtin_neon_vmlslu_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vmlsluv2si) \
VECTOR_INTRIN(__builtin_neon_vmlsluv4hi) \
VECTOR_INTRIN(__builtin_neon_vmlsluv8qi) \
VECTOR_INTRIN(__builtin_neon_vmls_nv2sf) \
VECTOR_INTRIN(__builtin_neon_vmls_nv2si) \
VECTOR_INTRIN(__builtin_neon_vmls_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vmls_nv4sf) \
VECTOR_INTRIN(__builtin_neon_vmls_nv4si) \
VECTOR_INTRIN(__builtin_neon_vmls_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vmlsv16qi) \
VECTOR_INTRIN(__builtin_neon_vmlsv2sf) \
VECTOR_INTRIN(__builtin_neon_vmlsv2si) \
VECTOR_INTRIN(__builtin_neon_vmlsv4hi) \
VECTOR_INTRIN(__builtin_neon_vmlsv4sf) \
VECTOR_INTRIN(__builtin_neon_vmlsv4si) \
VECTOR_INTRIN(__builtin_neon_vmlsv8hi) \
VECTOR_INTRIN(__builtin_neon_vmlsv8qi) \
VECTOR_INTRIN(__builtin_neon_vmovlsv2si) \
VECTOR_INTRIN(__builtin_neon_vmovlsv4hi) \
VECTOR_INTRIN(__builtin_neon_vmovlsv8qi) \
VECTOR_INTRIN(__builtin_neon_vmovluv2si) \
VECTOR_INTRIN(__builtin_neon_vmovluv4hi) \
VECTOR_INTRIN(__builtin_neon_vmovluv8qi) \
VECTOR_INTRIN(__builtin_neon_vmovnv2di) \
VECTOR_INTRIN(__builtin_neon_vmovnv4si) \
VECTOR_INTRIN(__builtin_neon_vmovnv8hi) \
VECTOR_INTRIN(__builtin_neon_vmulfv2sf) \
VECTOR_INTRIN(__builtin_neon_vmulfv4sf) \
VECTOR_INTRIN(__builtin_neon_vmul_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vmul_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vmul_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vmul_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vmul_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vmul_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vmullpv8qi) \
VECTOR_INTRIN(__builtin_neon_vmulls_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vmulls_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vmulls_nv2si) \
VECTOR_INTRIN(__builtin_neon_vmulls_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vmullsv2si) \
VECTOR_INTRIN(__builtin_neon_vmullsv4hi) \
VECTOR_INTRIN(__builtin_neon_vmullsv8qi) \
VECTOR_INTRIN(__builtin_neon_vmullu_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vmullu_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vmullu_nv2si) \
VECTOR_INTRIN(__builtin_neon_vmullu_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vmulluv2si) \
VECTOR_INTRIN(__builtin_neon_vmulluv4hi) \
VECTOR_INTRIN(__builtin_neon_vmulluv8qi) \
VECTOR_INTRIN(__builtin_neon_vmul_nv2sf) \
VECTOR_INTRIN(__builtin_neon_vmul_nv2si) \
VECTOR_INTRIN(__builtin_neon_vmul_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vmul_nv4sf) \
VECTOR_INTRIN(__builtin_neon_vmul_nv4si) \
VECTOR_INTRIN(__builtin_neon_vmul_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vmulpv16qi) \
VECTOR_INTRIN(__builtin_neon_vmulpv8qi) \
VECTOR_INTRIN(__builtin_neon_vmvnv16qi) \
VECTOR_INTRIN(__builtin_neon_vmvnv2si) \
VECTOR_INTRIN(__builtin_neon_vmvnv4hi) \
VECTOR_INTRIN(__builtin_neon_vmvnv4si) \
VECTOR_INTRIN(__builtin_neon_vmvnv8hi) \
VECTOR_INTRIN(__builtin_neon_vmvnv8qi) \
VECTOR_INTRIN(__builtin_neon_vnegv16qi) \
VECTOR_INTRIN(__builtin_neon_vnegv2sf) \
VECTOR_INTRIN(__builtin_neon_vnegv2si) \
VECTOR_INTRIN(__builtin_neon_vnegv4hi) \
VECTOR_INTRIN(__builtin_neon_vnegv4sf) \
VECTOR_INTRIN(__builtin_neon_vnegv4si) \
VECTOR_INTRIN(__builtin_neon_vnegv8hi) \
VECTOR_INTRIN(__builtin_neon_vnegv8qi) \
VECTOR_INTRIN(__builtin_neon_vpadalsv16qi) \
VECTOR_INTRIN(__builtin_neon_vpadalsv2si) \
VECTOR_INTRIN(__builtin_neon_vpadalsv4hi) \
VECTOR_INTRIN(__builtin_neon_vpadalsv4si) \
VECTOR_INTRIN(__builtin_neon_vpadalsv8hi) \
VECTOR_INTRIN(__builtin_neon_vpadalsv8qi) \
VECTOR_INTRIN(__builtin_neon_vpadaluv16qi) \
VECTOR_INTRIN(__builtin_neon_vpadaluv2si) \
VECTOR_INTRIN(__builtin_neon_vpadaluv4hi) \
VECTOR_INTRIN(__builtin_neon_vpadaluv4si) \
VECTOR_INTRIN(__builtin_neon_vpadaluv8hi) \
VECTOR_INTRIN(__builtin_neon_vpadaluv8qi) \
VECTOR_INTRIN(__builtin_neon_vpaddlsv16qi) \
VECTOR_INTRIN(__builtin_neon_vpaddlsv2si) \
VECTOR_INTRIN(__builtin_neon_vpaddlsv4hi) \
VECTOR_INTRIN(__builtin_neon_vpaddlsv4si) \
VECTOR_INTRIN(__builtin_neon_vpaddlsv8hi) \
VECTOR_INTRIN(__builtin_neon_vpaddlsv8qi) \
VECTOR_INTRIN(__builtin_neon_vpaddluv16qi) \
VECTOR_INTRIN(__builtin_neon_vpaddluv2si) \
VECTOR_INTRIN(__builtin_neon_vpaddluv4hi) \
VECTOR_INTRIN(__builtin_neon_vpaddluv4si) \
VECTOR_INTRIN(__builtin_neon_vpaddluv8hi) \
VECTOR_INTRIN(__builtin_neon_vpaddluv8qi) \
VECTOR_INTRIN(__builtin_neon_vpaddv2sf) \
VECTOR_INTRIN(__builtin_neon_vpaddv2si) \
VECTOR_INTRIN(__builtin_neon_vpaddv4hi) \
VECTOR_INTRIN(__builtin_neon_vpaddv8qi) \
VECTOR_INTRIN(__builtin_neon_vpmaxfv2sf) \
VECTOR_INTRIN(__builtin_neon_vpmaxsv2si) \
VECTOR_INTRIN(__builtin_neon_vpmaxsv4hi) \
VECTOR_INTRIN(__builtin_neon_vpmaxsv8qi) \
VECTOR_INTRIN(__builtin_neon_vpmaxuv2si) \
VECTOR_INTRIN(__builtin_neon_vpmaxuv4hi) \
VECTOR_INTRIN(__builtin_neon_vpmaxuv8qi) \
VECTOR_INTRIN(__builtin_neon_vpminfv2sf) \
VECTOR_INTRIN(__builtin_neon_vpminsv2si) \
VECTOR_INTRIN(__builtin_neon_vpminsv4hi) \
VECTOR_INTRIN(__builtin_neon_vpminsv8qi) \
VECTOR_INTRIN(__builtin_neon_vpminuv2si) \
VECTOR_INTRIN(__builtin_neon_vpminuv4hi) \
VECTOR_INTRIN(__builtin_neon_vpminuv8qi) \
VECTOR_INTRIN(__builtin_neon_vqabsv16qi) \
VECTOR_INTRIN(__builtin_neon_vqabsv2si) \
VECTOR_INTRIN(__builtin_neon_vqabsv4hi) \
VECTOR_INTRIN(__builtin_neon_vqabsv4si) \
VECTOR_INTRIN(__builtin_neon_vqabsv8hi) \
VECTOR_INTRIN(__builtin_neon_vqabsv8qi) \
VECTOR_INTRIN(__builtin_neon_vqaddsdi) \
VECTOR_INTRIN(__builtin_neon_vqaddsv16qi) \
VECTOR_INTRIN(__builtin_neon_vqaddsv2di) \
VECTOR_INTRIN(__builtin_neon_vqaddsv2si) \
VECTOR_INTRIN(__builtin_neon_vqaddsv4hi) \
VECTOR_INTRIN(__builtin_neon_vqaddsv4si) \
VECTOR_INTRIN(__builtin_neon_vqaddsv8hi) \
VECTOR_INTRIN(__builtin_neon_vqaddsv8qi) \
VECTOR_INTRIN(__builtin_neon_vqaddudi) \
VECTOR_INTRIN(__builtin_neon_vqadduv16qi) \
VECTOR_INTRIN(__builtin_neon_vqadduv2di) \
VECTOR_INTRIN(__builtin_neon_vqadduv2si) \
VECTOR_INTRIN(__builtin_neon_vqadduv4hi) \
VECTOR_INTRIN(__builtin_neon_vqadduv4si) \
VECTOR_INTRIN(__builtin_neon_vqadduv8hi) \
VECTOR_INTRIN(__builtin_neon_vqadduv8qi) \
VECTOR_INTRIN(__builtin_neon_vqdmlal_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vqdmlal_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmlal_nv2si) \
VECTOR_INTRIN(__builtin_neon_vqdmlal_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmlalv2si) \
VECTOR_INTRIN(__builtin_neon_vqdmlalv4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmlsl_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vqdmlsl_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmlsl_nv2si) \
VECTOR_INTRIN(__builtin_neon_vqdmlsl_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmlslv2si) \
VECTOR_INTRIN(__builtin_neon_vqdmlslv4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmulh_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vqdmulh_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmulh_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vqdmulh_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vqdmulh_nv2si) \
VECTOR_INTRIN(__builtin_neon_vqdmulh_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmulh_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqdmulh_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqdmulhv2si) \
VECTOR_INTRIN(__builtin_neon_vqdmulhv4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmulhv4si) \
VECTOR_INTRIN(__builtin_neon_vqdmulhv8hi) \
VECTOR_INTRIN(__builtin_neon_vqdmull_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vqdmull_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmull_nv2si) \
VECTOR_INTRIN(__builtin_neon_vqdmull_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vqdmullv2si) \
VECTOR_INTRIN(__builtin_neon_vqdmullv4hi) \
VECTOR_INTRIN(__builtin_neon_vqmovnsv2di) \
VECTOR_INTRIN(__builtin_neon_vqmovnsv4si) \
VECTOR_INTRIN(__builtin_neon_vqmovnsv8hi) \
VECTOR_INTRIN(__builtin_neon_vqmovnuv2di) \
VECTOR_INTRIN(__builtin_neon_vqmovnuv4si) \
VECTOR_INTRIN(__builtin_neon_vqmovnuv8hi) \
VECTOR_INTRIN(__builtin_neon_vqmovunv2di) \
VECTOR_INTRIN(__builtin_neon_vqmovunv4si) \
VECTOR_INTRIN(__builtin_neon_vqmovunv8hi) \
VECTOR_INTRIN(__builtin_neon_vqnegv16qi) \
VECTOR_INTRIN(__builtin_neon_vqnegv2si) \
VECTOR_INTRIN(__builtin_neon_vqnegv4hi) \
VECTOR_INTRIN(__builtin_neon_vqnegv4si) \
VECTOR_INTRIN(__builtin_neon_vqnegv8hi) \
VECTOR_INTRIN(__builtin_neon_vqnegv8qi) \
VECTOR_INTRIN(__builtin_neon_vqrdmulh_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vqrdmulh_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vqrdmulh_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vqrdmulh_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vqrdmulh_nv2si) \
VECTOR_INTRIN(__builtin_neon_vqrdmulh_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vqrdmulh_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqrdmulh_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqrdmulhv2si) \
VECTOR_INTRIN(__builtin_neon_vqrdmulhv4hi) \
VECTOR_INTRIN(__builtin_neon_vqrdmulhv4si) \
VECTOR_INTRIN(__builtin_neon_vqrdmulhv8hi) \
VECTOR_INTRIN(__builtin_neon_vqrshlsdi) \
VECTOR_INTRIN(__builtin_neon_vqrshlsv16qi) \
VECTOR_INTRIN(__builtin_neon_vqrshlsv2di) \
VECTOR_INTRIN(__builtin_neon_vqrshlsv2si) \
VECTOR_INTRIN(__builtin_neon_vqrshlsv4hi) \
VECTOR_INTRIN(__builtin_neon_vqrshlsv4si) \
VECTOR_INTRIN(__builtin_neon_vqrshlsv8hi) \
VECTOR_INTRIN(__builtin_neon_vqrshlsv8qi) \
VECTOR_INTRIN(__builtin_neon_vqrshludi) \
VECTOR_INTRIN(__builtin_neon_vqrshluv16qi) \
VECTOR_INTRIN(__builtin_neon_vqrshluv2di) \
VECTOR_INTRIN(__builtin_neon_vqrshluv2si) \
VECTOR_INTRIN(__builtin_neon_vqrshluv4hi) \
VECTOR_INTRIN(__builtin_neon_vqrshluv4si) \
VECTOR_INTRIN(__builtin_neon_vqrshluv8hi) \
VECTOR_INTRIN(__builtin_neon_vqrshluv8qi) \
VECTOR_INTRIN(__builtin_neon_vqrshrns_nv2di) \
VECTOR_INTRIN(__builtin_neon_vqrshrns_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqrshrns_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqrshrnu_nv2di) \
VECTOR_INTRIN(__builtin_neon_vqrshrnu_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqrshrnu_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqrshrun_nv2di) \
VECTOR_INTRIN(__builtin_neon_vqrshrun_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqrshrun_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqshlsdi) \
VECTOR_INTRIN(__builtin_neon_vqshl_s_ndi) \
VECTOR_INTRIN(__builtin_neon_vqshl_s_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vqshl_s_nv2di) \
VECTOR_INTRIN(__builtin_neon_vqshl_s_nv2si) \
VECTOR_INTRIN(__builtin_neon_vqshl_s_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vqshl_s_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqshl_s_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqshl_s_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vqshlsv16qi) \
VECTOR_INTRIN(__builtin_neon_vqshlsv2di) \
VECTOR_INTRIN(__builtin_neon_vqshlsv2si) \
VECTOR_INTRIN(__builtin_neon_vqshlsv4hi) \
VECTOR_INTRIN(__builtin_neon_vqshlsv4si) \
VECTOR_INTRIN(__builtin_neon_vqshlsv8hi) \
VECTOR_INTRIN(__builtin_neon_vqshlsv8qi) \
VECTOR_INTRIN(__builtin_neon_vqshludi) \
VECTOR_INTRIN(__builtin_neon_vqshl_u_ndi) \
VECTOR_INTRIN(__builtin_neon_vqshlu_ndi) \
VECTOR_INTRIN(__builtin_neon_vqshl_u_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vqshlu_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vqshl_u_nv2di) \
VECTOR_INTRIN(__builtin_neon_vqshlu_nv2di) \
VECTOR_INTRIN(__builtin_neon_vqshl_u_nv2si) \
VECTOR_INTRIN(__builtin_neon_vqshlu_nv2si) \
VECTOR_INTRIN(__builtin_neon_vqshl_u_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vqshlu_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vqshl_u_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqshlu_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqshl_u_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqshlu_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqshl_u_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vqshlu_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vqshluv16qi) \
VECTOR_INTRIN(__builtin_neon_vqshluv2di) \
VECTOR_INTRIN(__builtin_neon_vqshluv2si) \
VECTOR_INTRIN(__builtin_neon_vqshluv4hi) \
VECTOR_INTRIN(__builtin_neon_vqshluv4si) \
VECTOR_INTRIN(__builtin_neon_vqshluv8hi) \
VECTOR_INTRIN(__builtin_neon_vqshluv8qi) \
VECTOR_INTRIN(__builtin_neon_vqshrns_nv2di) \
VECTOR_INTRIN(__builtin_neon_vqshrns_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqshrns_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqshrnu_nv2di) \
VECTOR_INTRIN(__builtin_neon_vqshrnu_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqshrnu_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqshrun_nv2di) \
VECTOR_INTRIN(__builtin_neon_vqshrun_nv4si) \
VECTOR_INTRIN(__builtin_neon_vqshrun_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vqsubsdi) \
VECTOR_INTRIN(__builtin_neon_vqsubsv16qi) \
VECTOR_INTRIN(__builtin_neon_vqsubsv2di) \
VECTOR_INTRIN(__builtin_neon_vqsubsv2si) \
VECTOR_INTRIN(__builtin_neon_vqsubsv4hi) \
VECTOR_INTRIN(__builtin_neon_vqsubsv4si) \
VECTOR_INTRIN(__builtin_neon_vqsubsv8hi) \
VECTOR_INTRIN(__builtin_neon_vqsubsv8qi) \
VECTOR_INTRIN(__builtin_neon_vqsubudi) \
VECTOR_INTRIN(__builtin_neon_vqsubuv16qi) \
VECTOR_INTRIN(__builtin_neon_vqsubuv2di) \
VECTOR_INTRIN(__builtin_neon_vqsubuv2si) \
VECTOR_INTRIN(__builtin_neon_vqsubuv4hi) \
VECTOR_INTRIN(__builtin_neon_vqsubuv4si) \
VECTOR_INTRIN(__builtin_neon_vqsubuv8hi) \
VECTOR_INTRIN(__builtin_neon_vqsubuv8qi) \
VECTOR_INTRIN(__builtin_neon_vraddhnv2di) \
VECTOR_INTRIN(__builtin_neon_vraddhnv4si) \
VECTOR_INTRIN(__builtin_neon_vraddhnv8hi) \
VECTOR_INTRIN(__builtin_neon_vrecpev2sf) \
VECTOR_INTRIN(__builtin_neon_vrecpev2si) \
VECTOR_INTRIN(__builtin_neon_vrecpev4sf) \
VECTOR_INTRIN(__builtin_neon_vrecpev4si) \
VECTOR_INTRIN(__builtin_neon_vrecpsv2sf) \
VECTOR_INTRIN(__builtin_neon_vrecpsv4sf) \
VECTOR_INTRIN(__builtin_neon_vreinterpretdidi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretdiv2sf) \
VECTOR_INTRIN(__builtin_neon_vreinterpretdiv2si) \
VECTOR_INTRIN(__builtin_neon_vreinterpretdiv4hi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretdiv8qi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv16qiv16qi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv16qiv2di) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv16qiv4sf) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv16qiv4si) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv16qiv8hi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2div16qi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2div2di) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2div4sf) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2div4si) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2div8hi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2sfdi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2sfv2si) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2sfv4hi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2sfv8qi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2sidi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2siv2sf) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2siv2si) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2siv4hi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv2siv8qi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4hidi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4hiv2sf) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4hiv2si) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4hiv4hi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4hiv8qi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4sfv16qi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4sfv2di) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4sfv4si) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4sfv8hi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4siv16qi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4siv2di) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4siv4sf) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4siv4si) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv4siv8hi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv8hiv16qi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv8hiv2di) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv8hiv4sf) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv8hiv4si) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv8hiv8hi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv8qidi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv8qiv2sf) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv8qiv2si) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv8qiv4hi) \
VECTOR_INTRIN(__builtin_neon_vreinterpretv8qiv8qi) \
VECTOR_INTRIN(__builtin_neon_vrhaddsv16qi) \
VECTOR_INTRIN(__builtin_neon_vrhaddsv2si) \
VECTOR_INTRIN(__builtin_neon_vrhaddsv4hi) \
VECTOR_INTRIN(__builtin_neon_vrhaddsv4si) \
VECTOR_INTRIN(__builtin_neon_vrhaddsv8hi) \
VECTOR_INTRIN(__builtin_neon_vrhaddsv8qi) \
VECTOR_INTRIN(__builtin_neon_vrhadduv16qi) \
VECTOR_INTRIN(__builtin_neon_vrhadduv2si) \
VECTOR_INTRIN(__builtin_neon_vrhadduv4hi) \
VECTOR_INTRIN(__builtin_neon_vrhadduv4si) \
VECTOR_INTRIN(__builtin_neon_vrhadduv8hi) \
VECTOR_INTRIN(__builtin_neon_vrhadduv8qi) \
VECTOR_INTRIN(__builtin_neon_vrshlsdi) \
VECTOR_INTRIN(__builtin_neon_vrshlsv16qi) \
VECTOR_INTRIN(__builtin_neon_vrshlsv2di) \
VECTOR_INTRIN(__builtin_neon_vrshlsv2si) \
VECTOR_INTRIN(__builtin_neon_vrshlsv4hi) \
VECTOR_INTRIN(__builtin_neon_vrshlsv4si) \
VECTOR_INTRIN(__builtin_neon_vrshlsv8hi) \
VECTOR_INTRIN(__builtin_neon_vrshlsv8qi) \
VECTOR_INTRIN(__builtin_neon_vrshludi) \
VECTOR_INTRIN(__builtin_neon_vrshluv16qi) \
VECTOR_INTRIN(__builtin_neon_vrshluv2di) \
VECTOR_INTRIN(__builtin_neon_vrshluv2si) \
VECTOR_INTRIN(__builtin_neon_vrshluv4hi) \
VECTOR_INTRIN(__builtin_neon_vrshluv4si) \
VECTOR_INTRIN(__builtin_neon_vrshluv8hi) \
VECTOR_INTRIN(__builtin_neon_vrshluv8qi) \
VECTOR_INTRIN(__builtin_neon_vrshrn_nv2di) \
VECTOR_INTRIN(__builtin_neon_vrshrn_nv4si) \
VECTOR_INTRIN(__builtin_neon_vrshrn_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vrshrs_ndi) \
VECTOR_INTRIN(__builtin_neon_vrshrs_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vrshrs_nv2di) \
VECTOR_INTRIN(__builtin_neon_vrshrs_nv2si) \
VECTOR_INTRIN(__builtin_neon_vrshrs_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vrshrs_nv4si) \
VECTOR_INTRIN(__builtin_neon_vrshrs_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vrshrs_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vrshru_ndi) \
VECTOR_INTRIN(__builtin_neon_vrshru_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vrshru_nv2di) \
VECTOR_INTRIN(__builtin_neon_vrshru_nv2si) \
VECTOR_INTRIN(__builtin_neon_vrshru_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vrshru_nv4si) \
VECTOR_INTRIN(__builtin_neon_vrshru_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vrshru_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vrsqrtev2sf) \
VECTOR_INTRIN(__builtin_neon_vrsqrtev2si) \
VECTOR_INTRIN(__builtin_neon_vrsqrtev4sf) \
VECTOR_INTRIN(__builtin_neon_vrsqrtev4si) \
VECTOR_INTRIN(__builtin_neon_vrsqrtsv2sf) \
VECTOR_INTRIN(__builtin_neon_vrsqrtsv4sf) \
VECTOR_INTRIN(__builtin_neon_vrsras_ndi) \
VECTOR_INTRIN(__builtin_neon_vrsras_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vrsras_nv2di) \
VECTOR_INTRIN(__builtin_neon_vrsras_nv2si) \
VECTOR_INTRIN(__builtin_neon_vrsras_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vrsras_nv4si) \
VECTOR_INTRIN(__builtin_neon_vrsras_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vrsras_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vrsrau_ndi) \
VECTOR_INTRIN(__builtin_neon_vrsrau_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vrsrau_nv2di) \
VECTOR_INTRIN(__builtin_neon_vrsrau_nv2si) \
VECTOR_INTRIN(__builtin_neon_vrsrau_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vrsrau_nv4si) \
VECTOR_INTRIN(__builtin_neon_vrsrau_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vrsrau_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vrsubhnv2di) \
VECTOR_INTRIN(__builtin_neon_vrsubhnv4si) \
VECTOR_INTRIN(__builtin_neon_vrsubhnv8hi) \
VECTOR_INTRIN(__builtin_neon_vset_lanedi) \
VECTOR_INTRIN(__builtin_neon_vset_lanev16qi) \
VECTOR_INTRIN(__builtin_neon_vset_lanev2di) \
VECTOR_INTRIN(__builtin_neon_vset_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vset_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vset_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vset_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vset_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vset_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vset_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vshlls_nv2si) \
VECTOR_INTRIN(__builtin_neon_vshlls_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vshlls_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vshllu_nv2si) \
VECTOR_INTRIN(__builtin_neon_vshllu_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vshllu_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vshl_ndi) \
VECTOR_INTRIN(__builtin_neon_vshl_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vshl_nv2di) \
VECTOR_INTRIN(__builtin_neon_vshl_nv2si) \
VECTOR_INTRIN(__builtin_neon_vshl_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vshl_nv4si) \
VECTOR_INTRIN(__builtin_neon_vshl_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vshl_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vshlsdi) \
VECTOR_INTRIN(__builtin_neon_vshlsv16qi) \
VECTOR_INTRIN(__builtin_neon_vshlsv2di) \
VECTOR_INTRIN(__builtin_neon_vshlsv2si) \
VECTOR_INTRIN(__builtin_neon_vshlsv4hi) \
VECTOR_INTRIN(__builtin_neon_vshlsv4si) \
VECTOR_INTRIN(__builtin_neon_vshlsv8hi) \
VECTOR_INTRIN(__builtin_neon_vshlsv8qi) \
VECTOR_INTRIN(__builtin_neon_vshludi) \
VECTOR_INTRIN(__builtin_neon_vshluv16qi) \
VECTOR_INTRIN(__builtin_neon_vshluv2di) \
VECTOR_INTRIN(__builtin_neon_vshluv2si) \
VECTOR_INTRIN(__builtin_neon_vshluv4hi) \
VECTOR_INTRIN(__builtin_neon_vshluv4si) \
VECTOR_INTRIN(__builtin_neon_vshluv8hi) \
VECTOR_INTRIN(__builtin_neon_vshluv8qi) \
VECTOR_INTRIN(__builtin_neon_vshrn_nv2di) \
VECTOR_INTRIN(__builtin_neon_vshrn_nv4si) \
VECTOR_INTRIN(__builtin_neon_vshrn_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vshrs_ndi) \
VECTOR_INTRIN(__builtin_neon_vshrs_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vshrs_nv2di) \
VECTOR_INTRIN(__builtin_neon_vshrs_nv2si) \
VECTOR_INTRIN(__builtin_neon_vshrs_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vshrs_nv4si) \
VECTOR_INTRIN(__builtin_neon_vshrs_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vshrs_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vshru_ndi) \
VECTOR_INTRIN(__builtin_neon_vshru_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vshru_nv2di) \
VECTOR_INTRIN(__builtin_neon_vshru_nv2si) \
VECTOR_INTRIN(__builtin_neon_vshru_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vshru_nv4si) \
VECTOR_INTRIN(__builtin_neon_vshru_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vshru_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vsli_ndi) \
VECTOR_INTRIN(__builtin_neon_vsli_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vsli_nv2di) \
VECTOR_INTRIN(__builtin_neon_vsli_nv2si) \
VECTOR_INTRIN(__builtin_neon_vsli_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vsli_nv4si) \
VECTOR_INTRIN(__builtin_neon_vsli_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vsli_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vsras_ndi) \
VECTOR_INTRIN(__builtin_neon_vsras_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vsras_nv2di) \
VECTOR_INTRIN(__builtin_neon_vsras_nv2si) \
VECTOR_INTRIN(__builtin_neon_vsras_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vsras_nv4si) \
VECTOR_INTRIN(__builtin_neon_vsras_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vsras_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vsrau_ndi) \
VECTOR_INTRIN(__builtin_neon_vsrau_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vsrau_nv2di) \
VECTOR_INTRIN(__builtin_neon_vsrau_nv2si) \
VECTOR_INTRIN(__builtin_neon_vsrau_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vsrau_nv4si) \
VECTOR_INTRIN(__builtin_neon_vsrau_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vsrau_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vsri_ndi) \
VECTOR_INTRIN(__builtin_neon_vsri_nv16qi) \
VECTOR_INTRIN(__builtin_neon_vsri_nv2di) \
VECTOR_INTRIN(__builtin_neon_vsri_nv2si) \
VECTOR_INTRIN(__builtin_neon_vsri_nv4hi) \
VECTOR_INTRIN(__builtin_neon_vsri_nv4si) \
VECTOR_INTRIN(__builtin_neon_vsri_nv8hi) \
VECTOR_INTRIN(__builtin_neon_vsri_nv8qi) \
VECTOR_INTRIN(__builtin_neon_vst1di) \
VECTOR_INTRIN(__builtin_neon_vst1_lanedi) \
VECTOR_INTRIN(__builtin_neon_vst1_lanev16qi) \
VECTOR_INTRIN(__builtin_neon_vst1_lanev2di) \
VECTOR_INTRIN(__builtin_neon_vst1_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vst1_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vst1_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vst1_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vst1_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vst1_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vst1_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vst1v16qi) \
VECTOR_INTRIN(__builtin_neon_vst1v2di) \
VECTOR_INTRIN(__builtin_neon_vst1v2sf) \
VECTOR_INTRIN(__builtin_neon_vst1v2si) \
VECTOR_INTRIN(__builtin_neon_vst1v4hi) \
VECTOR_INTRIN(__builtin_neon_vst1v4sf) \
VECTOR_INTRIN(__builtin_neon_vst1v4si) \
VECTOR_INTRIN(__builtin_neon_vst1v8hi) \
VECTOR_INTRIN(__builtin_neon_vst1v8qi) \
VECTOR_INTRIN(__builtin_neon_vst2di) \
VECTOR_INTRIN(__builtin_neon_vst2_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vst2_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vst2_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vst2_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vst2_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vst2_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vst2_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vst2v16qi) \
VECTOR_INTRIN(__builtin_neon_vst2v2sf) \
VECTOR_INTRIN(__builtin_neon_vst2v2si) \
VECTOR_INTRIN(__builtin_neon_vst2v4hi) \
VECTOR_INTRIN(__builtin_neon_vst2v4sf) \
VECTOR_INTRIN(__builtin_neon_vst2v4si) \
VECTOR_INTRIN(__builtin_neon_vst2v8hi) \
VECTOR_INTRIN(__builtin_neon_vst2v8qi) \
VECTOR_INTRIN(__builtin_neon_vst3di) \
VECTOR_INTRIN(__builtin_neon_vst3_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vst3_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vst3_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vst3_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vst3_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vst3_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vst3_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vst3v16qi) \
VECTOR_INTRIN(__builtin_neon_vst3v2sf) \
VECTOR_INTRIN(__builtin_neon_vst3v2si) \
VECTOR_INTRIN(__builtin_neon_vst3v4hi) \
VECTOR_INTRIN(__builtin_neon_vst3v4sf) \
VECTOR_INTRIN(__builtin_neon_vst3v4si) \
VECTOR_INTRIN(__builtin_neon_vst3v8hi) \
VECTOR_INTRIN(__builtin_neon_vst3v8qi) \
VECTOR_INTRIN(__builtin_neon_vst4di) \
VECTOR_INTRIN(__builtin_neon_vst4_lanev2sf) \
VECTOR_INTRIN(__builtin_neon_vst4_lanev2si) \
VECTOR_INTRIN(__builtin_neon_vst4_lanev4hi) \
VECTOR_INTRIN(__builtin_neon_vst4_lanev4sf) \
VECTOR_INTRIN(__builtin_neon_vst4_lanev4si) \
VECTOR_INTRIN(__builtin_neon_vst4_lanev8hi) \
VECTOR_INTRIN(__builtin_neon_vst4_lanev8qi) \
VECTOR_INTRIN(__builtin_neon_vst4v16qi) \
VECTOR_INTRIN(__builtin_neon_vst4v2sf) \
VECTOR_INTRIN(__builtin_neon_vst4v2si) \
VECTOR_INTRIN(__builtin_neon_vst4v4hi) \
VECTOR_INTRIN(__builtin_neon_vst4v4sf) \
VECTOR_INTRIN(__builtin_neon_vst4v4si) \
VECTOR_INTRIN(__builtin_neon_vst4v8hi) \
VECTOR_INTRIN(__builtin_neon_vst4v8qi) \
VECTOR_INTRIN(__builtin_neon_vsubhnv2di) \
VECTOR_INTRIN(__builtin_neon_vsubhnv4si) \
VECTOR_INTRIN(__builtin_neon_vsubhnv8hi) \
VECTOR_INTRIN(__builtin_neon_vsublsv2si) \
VECTOR_INTRIN(__builtin_neon_vsublsv4hi) \
VECTOR_INTRIN(__builtin_neon_vsublsv8qi) \
VECTOR_INTRIN(__builtin_neon_vsubluv2si) \
VECTOR_INTRIN(__builtin_neon_vsubluv4hi) \
VECTOR_INTRIN(__builtin_neon_vsubluv8qi) \
VECTOR_INTRIN(__builtin_neon_vsubv2sf) \
VECTOR_INTRIN(__builtin_neon_vsubv4sf) \
VECTOR_INTRIN(__builtin_neon_vsubwsv2si) \
VECTOR_INTRIN(__builtin_neon_vsubwsv4hi) \
VECTOR_INTRIN(__builtin_neon_vsubwsv8qi) \
VECTOR_INTRIN(__builtin_neon_vsubwuv2si) \
VECTOR_INTRIN(__builtin_neon_vsubwuv4hi) \
VECTOR_INTRIN(__builtin_neon_vsubwuv8qi) \
VECTOR_INTRIN(__builtin_neon_vtbl1v8qi) \
VECTOR_INTRIN(__builtin_neon_vtbl2v8qi) \
VECTOR_INTRIN(__builtin_neon_vtbl3v8qi) \
VECTOR_INTRIN(__builtin_neon_vtbl4v8qi) \
VECTOR_INTRIN(__builtin_neon_vtbx1v8qi) \
VECTOR_INTRIN(__builtin_neon_vtbx2v8qi) \
VECTOR_INTRIN(__builtin_neon_vtbx3v8qi) \
VECTOR_INTRIN(__builtin_neon_vtbx4v8qi) \
VECTOR_INTRIN(__builtin_neon_vtstv16qi) \
VECTOR_INTRIN(__builtin_neon_vtstv2si) \
VECTOR_INTRIN(__builtin_neon_vtstv4hi) \
VECTOR_INTRIN(__builtin_neon_vtstv4si) \
VECTOR_INTRIN(__builtin_neon_vtstv8hi) \
VECTOR_INTRIN(__builtin_neon_vtstv8qi) \
END

// VECTOR_INTRIN(__builtin_shuffle) \

static void do_alias(const char* newname, const char* existing)
{
    std::cout << "{\n"
        << "scope_entry_list_t *entry_list = query_in_scope_str(decl_context, uniquestr(\"" << existing << "\"), /* field_path */ NULL);\n"
        << "ERROR_CONDITION(entry_list == NULL, \"Symbol '" << existing << "' should have been declared\",0);\n"
        << "scope_entry_t* orig_sym = entry_list_head(entry_list);\n"
        << "entry_list_free(entry_list);\n"
        << "scope_entry_t* new_sym = new_symbol(decl_context, decl_context->current_scope, uniquestr(\"" << newname << "\"));\n"
        << "new_sym->kind = SK_FUNCTION;"
        << "new_sym->do_not_print = 1;\n"
        << "new_sym->type_information = orig_sym->type_information;\n"
        << "symbol_entity_specs_set_is_builtin(new_sym, 1);\n"
        << "}\n";
}

int main(int, char**)
{
#define VECTOR_INTRIN(X) \
    f<__typeof__(X)>(#X);
#define VECTOR_ALIAS(newname, existing) \
    do_alias(#newname, #existing);
    VECTOR_INTRINSICS_LIST
#undef VECTOR_INTRIN
}
